@@ -7,8 +7,11 @@
 #include "qaudiodevice.h"
 #include "qaudiosink.h"
 #include "qmediadevices.h"
+#include "qaudiobuffer.h"
 #include <QtCore/qloggingcategory.h>
 #include <private/qplatformmediadevices_p.h>
+#include <private/qplatformmediaintegration_p.h>
+#include <private/qplatformaudioresampler_p.h>
 
 static Q_LOGGING_CATEGORY(qLcSoundEffect, "qt.multimedia.soundeffect")
 
@@ -16,6 +19,27 @@ QT_BEGIN_NAMESPACE
 
 Q_GLOBAL_STATIC(QSampleCache, sampleCache)
 
+namespace
+{
+struct AudioSinkDeleter
+{
+    void operator ()(QAudioSink* sink) const
+    {
+        sink->stop();
+        // Investigate:should we just delete?
+        sink->deleteLater();
+    }
+};
+
+struct SampleDeleter
+{
+    void operator ()(QSample* sample) const
+    {
+        sample->release();
+    }
+};
+}
+
 class QSoundEffectPrivate : public QIODevice
 {
 public:
@@ -27,13 +51,14 @@ public:
     qint64 size() const override {
         if (m_sample->state() != QSample::Ready)
             return 0;
-        return m_loopCount == QSoundEffect::Infinite ? 0 : m_loopCount * m_sample->data().size();
+        return m_loopCount == QSoundEffect::Infinite ? 0 : m_loopCount * m_audioBuffer.byteCount();
     }
     qint64 bytesAvailable() const override {
         if (m_sample->state() != QSample::Ready)
             return 0;
-        return m_loopCount == QSoundEffect::Infinite
-                   ? std::numeric_limits<qint64>::max() : m_runningCount * m_sample->data().size() - m_offset;
+        if (m_loopCount == QSoundEffect::Infinite)
+            return std::numeric_limits<qint64>::max();
+        return m_runningCount * m_audioBuffer.byteCount() - m_offset;
     }
     bool isSequential() const override {
         return m_loopCount == QSoundEffect::Infinite;
@@ -45,10 +70,11 @@ public:
     void setLoopsRemaining(int loopsRemaining);
     void setStatus(QSoundEffect::Status status);
     void setPlaying(bool playing);
+    bool updateAudioOutput();
 
 public Q_SLOTS:
-    void sampleReady();
-    void decoderError();
+    void sampleReady(QSample *);
+    void decoderError(QSample *);
     void stateChanged(QAudio::State);
 
 public:
@@ -57,9 +83,10 @@ public:
     int m_loopCount = 1;
     int m_runningCount = 0;
     bool m_playing = false;
-    QSoundEffect::Status  m_status = QSoundEffect::Null;
-    QAudioSink *m_audioOutput = nullptr;
-    QSample *m_sample = nullptr;
+    QSoundEffect::Status m_status = QSoundEffect::Null;
+    std::unique_ptr<QAudioSink, AudioSinkDeleter> m_audioSink;
+    std::unique_ptr<QSample, SampleDeleter> m_sample;
+    QAudioBuffer m_audioBuffer;
     bool m_muted = false;
     float m_volume = 1.0;
     bool m_sampleReady = false;
@@ -74,39 +101,43 @@ QSoundEffectPrivate::QSoundEffectPrivate(QSoundEffect *q, const QAudioDevice &au
 {
     open(QIODevice::ReadOnly);
 
-    QPlatformMediaDevices::instance()->prepareAudio();
+    QPlatformMediaIntegration::instance()->mediaDevices()->prepareAudio();
 }
 
-void QSoundEffectPrivate::sampleReady()
+void QSoundEffectPrivate::sampleReady(QSample *sample)
 {
+    if (sample && sample != m_sample.get())
+        return;
+
     if (m_status == QSoundEffect::Error)
         return;
 
     qCDebug(qLcSoundEffect) << this << "sampleReady: sample size:" << m_sample->data().size();
-    disconnect(m_sample, &QSample::error, this, &QSoundEffectPrivate::decoderError);
-    disconnect(m_sample, &QSample::ready, this, &QSoundEffectPrivate::sampleReady);
-    if (!m_audioOutput) {
-        m_audioOutput = new QAudioSink(m_audioDevice, m_sample->format());
-        connect(m_audioOutput, &QAudioSink::stateChanged, this, &QSoundEffectPrivate::stateChanged);
-        if (!m_muted)
-            m_audioOutput->setVolume(m_volume);
-        else
-            m_audioOutput->setVolume(0);
+    disconnect(m_sample.get(), &QSample::error, this, &QSoundEffectPrivate::decoderError);
+    disconnect(m_sample.get(), &QSample::ready, this, &QSoundEffectPrivate::sampleReady);
+
+    if (!m_audioSink) {
+        if (!updateAudioOutput()) // Create audio sink
+            return; // Returns if no audio devices are available
     }
+
     m_sampleReady = true;
     setStatus(QSoundEffect::Ready);
 
-    if (m_playing && m_audioOutput->state() == QAudio::StoppedState) {
+    if (m_playing && m_audioSink->state() == QAudio::StoppedState) {
         qCDebug(qLcSoundEffect) << this << "starting playback on audiooutput";
-        m_audioOutput->start(this);
+        m_audioSink->start(this);
     }
 }
 
-void QSoundEffectPrivate::decoderError()
+void QSoundEffectPrivate::decoderError(QSample *sample)
 {
+    if (sample && sample != m_sample.get())
+        return;
+
     qWarning("QSoundEffect(qaudio): Error decoding source %ls", qUtf16Printable(m_url.toString()));
-    disconnect(m_sample, &QSample::ready, this, &QSoundEffectPrivate::sampleReady);
-    disconnect(m_sample, &QSample::error, this, &QSoundEffectPrivate::decoderError);
+    disconnect(m_sample.get(), &QSample::ready, this, &QSoundEffectPrivate::sampleReady);
+    disconnect(m_sample.get(), &QSample::error, this, &QSoundEffectPrivate::decoderError);
     m_playing = false;
     setStatus(QSoundEffect::Error);
 }
@@ -118,6 +149,63 @@ void QSoundEffectPrivate::stateChanged(QAudio::State state)
         q_ptr->stop();
 }
 
+bool QSoundEffectPrivate::updateAudioOutput()
+{
+    const auto audioDevice =
+            m_audioDevice.isNull() ? QMediaDevices::defaultAudioOutput() : m_audioDevice;
+
+    if (audioDevice.isNull()) {
+        // We are likely on a virtual machine, for example in CI
+        qCCritical(qLcSoundEffect) << "Failed to update audio output. No audio devices available.";
+        setStatus(QSoundEffect::Error);
+        return false;
+    }
+
+    if (m_audioDevice.isNull()) {
+        q_ptr->setAudioDevice(audioDevice); // Updates m_audioDevice and emits audioDeviceChanged
+    }
+
+    m_audioBuffer = {};
+
+    Q_ASSERT(m_sample);
+
+    const auto &sampleFormat = m_sample->format();
+    const auto sampleChannelConfig =
+            sampleFormat.channelConfig() == QAudioFormat::ChannelConfigUnknown
+            ? QAudioFormat::defaultChannelConfigForChannelCount(sampleFormat.channelCount())
+            : sampleFormat.channelConfig();
+
+    if (sampleChannelConfig != audioDevice.channelConfiguration()
+        && audioDevice.channelConfiguration() != QAudioFormat::ChannelConfigUnknown) {
+        qCDebug(qLcSoundEffect) << "Create resampler for channels mapping: config"
+                                << sampleFormat.channelConfig() << "=> config"
+                                << audioDevice.channelConfiguration();
+        auto outputFormat = sampleFormat;
+        outputFormat.setChannelConfig(audioDevice.channelConfiguration());
+
+        const auto resampler = QPlatformMediaIntegration::instance()->createAudioResampler(
+                m_sample->format(), outputFormat);
+        if (resampler)
+            m_audioBuffer = resampler.value()->resample(m_sample->data().constData(),
+                                                        m_sample->data().size());
+        else
+            qCDebug(qLcSoundEffect) << "Cannot create resampler for channels mapping";
+    }
+
+    if (!m_audioBuffer.isValid())
+        m_audioBuffer = QAudioBuffer(m_sample->data(), m_sample->format());
+
+    m_audioSink.reset(new QAudioSink(audioDevice, m_audioBuffer.format()));
+
+    connect(m_audioSink.get(), &QAudioSink::stateChanged, this, &QSoundEffectPrivate::stateChanged);
+    if (!m_muted)
+        m_audioSink->setVolume(m_volume);
+    else
+        m_audioSink->setVolume(0);
+
+    return true;
+}
+
 qint64 QSoundEffectPrivate::readData(char *data, qint64 len)
 {
     qCDebug(qLcSoundEffect) << this << "readData" << len << m_runningCount;
@@ -130,11 +218,11 @@ qint64 QSoundEffectPrivate::readData(char *data, qint64 len)
 
     qint64 bytesWritten = 0;
 
-    const int   sampleSize = m_sample->data().size();
-    const char* sampleData = m_sample->data().constData();
+    const int sampleSize = m_audioBuffer.byteCount();
+    const char *sampleData = m_audioBuffer.constData<char>();
 
     while (len && m_runningCount) {
-        int toWrite = qMin(sampleSize - m_offset, len);
+        int toWrite = qMax(0, qMin(sampleSize - m_offset, len));
         memcpy(data, sampleData + m_offset, toWrite);
         bytesWritten += toWrite;
         data += toWrite;
@@ -181,8 +269,8 @@ void QSoundEffectPrivate::setStatus(QSoundEffect::Status status)
 void QSoundEffectPrivate::setPlaying(bool playing)
 {
     qCDebug(qLcSoundEffect) << this << "setPlaying(" << playing << ")" << m_playing;
-    if (m_audioOutput) {
-        m_audioOutput->stop();
+    if (m_audioSink) {
+        m_audioSink->stop();
         if (playing && !m_sampleReady)
             return;
     }
@@ -191,8 +279,8 @@ void QSoundEffectPrivate::setPlaying(bool playing)
         return;
     m_playing = playing;
 
-    if (m_audioOutput && playing)
-        m_audioOutput->start(this);
+    if (m_audioSink && playing)
+        m_audioSink->start(this);
 
     emit q_ptr->playingChanged();
 }
@@ -231,7 +319,7 @@ void QSoundEffectPrivate::setPlaying(bool playing)
 
 /*!
     \qmltype SoundEffect
-    \instantiates QSoundEffect
+    \nativetype QSoundEffect
     \brief The SoundEffect type provides a way to play sound effects in QML.
 
     \inmodule QtMultimedia
@@ -283,11 +371,8 @@ QSoundEffect::QSoundEffect(const QAudioDevice &audioDevice, QObject *parent)
 QSoundEffect::~QSoundEffect()
 {
     stop();
-    if (d->m_audioOutput) {
-        d->m_audioOutput->stop();
-        d->m_audioOutput->deleteLater();
-        d->m_sample->release();
-    }
+    d->m_audioSink.reset();
+    d->m_sample.reset();
     delete d;
 }
 
@@ -343,7 +428,6 @@ void QSoundEffect::setSource(const QUrl &url)
     stop();
 
     d->m_url = url;
-
     d->m_sampleReady = false;
 
     if (url.isEmpty()) {
@@ -358,31 +442,28 @@ void QSoundEffect::setSource(const QUrl &url)
 
     if (d->m_sample) {
         if (!d->m_sampleReady) {
-            QObject::disconnect(d->m_sample, &QSample::error, d, &QSoundEffectPrivate::decoderError);
-            QObject::disconnect(d->m_sample, &QSample::ready, d, &QSoundEffectPrivate::sampleReady);
+            disconnect(d->m_sample.get(), &QSample::error, d, &QSoundEffectPrivate::decoderError);
+            disconnect(d->m_sample.get(), &QSample::ready, d, &QSoundEffectPrivate::sampleReady);
         }
-        d->m_sample->release();
-        d->m_sample = nullptr;
+        d->m_sample.reset();
     }
 
-    if (d->m_audioOutput) {
-        QObject::disconnect(d->m_audioOutput, &QAudioSink::stateChanged, d, &QSoundEffectPrivate::stateChanged);
-        d->m_audioOutput->stop();
-        d->m_audioOutput->deleteLater();
-        d->m_audioOutput = nullptr;
+    if (d->m_audioSink) {
+        disconnect(d->m_audioSink.get(), &QAudioSink::stateChanged, d, &QSoundEffectPrivate::stateChanged);
+        d->m_audioSink.reset();
     }
 
     d->setStatus(QSoundEffect::Loading);
-    d->m_sample = sampleCache()->requestSample(url);
-    QObject::connect(d->m_sample, &QSample::error, d, &QSoundEffectPrivate::decoderError);
-    QObject::connect(d->m_sample, &QSample::ready, d, &QSoundEffectPrivate::sampleReady);
+    d->m_sample.reset(sampleCache()->requestSample(url));
+    connect(d->m_sample.get(), &QSample::error, d, &QSoundEffectPrivate::decoderError);
+    connect(d->m_sample.get(), &QSample::ready, d, &QSoundEffectPrivate::sampleReady);
 
     switch (d->m_sample->state()) {
     case QSample::Ready:
-        d->sampleReady();
+        d->sampleReady(d->m_sample.get());
         break;
     case QSample::Error:
-        d->decoderError();
+        d->decoderError(d->m_sample.get());
         break;
     default:
         break;
@@ -465,10 +546,28 @@ QAudioDevice QSoundEffect::audioDevice()
 
 void QSoundEffect::setAudioDevice(const QAudioDevice &device)
 {
+    qCDebug(qLcSoundEffect) << this << "setAudioDevice:" << device.description();
+
     if (d->m_audioDevice == device)
         return;
-    // ### recreate the QAudioSink if needed
+
     d->m_audioDevice = device;
+
+    if (!d->m_sampleReady) {
+        emit audioDeviceChanged();
+        return; // The audio sink will be recreated later by QSoundEffect::sampleReady()
+    }
+
+    bool playing = d->m_playing;
+    std::chrono::microseconds current_time{d->m_audioBuffer.format().durationForBytes(d->m_offset)};
+
+    // Recreate the QAudioSink with the new audio device and current sample
+    if (d->updateAudioOutput() && playing) {
+        // Resume playback from current position
+        d->m_offset = d->m_audioBuffer.format().bytesForDuration(current_time.count());
+        d->setPlaying(true);
+    }
+
     emit audioDeviceChanged();
 }
 
@@ -502,7 +601,7 @@ int QSoundEffect::loopsRemaining() const
 
     UI volume controls should usually be scaled non-linearly. For example, using a logarithmic scale
     will produce linear changes in perceived loudness, which is what a user would normally expect
-    from a volume control. See \l {QAudio::convertVolume()}{convertVolume()}
+    from a volume control. See \l {QtAudio::convertVolume()}{convertVolume()}
     for more details.
 */
 /*!
@@ -516,8 +615,8 @@ int QSoundEffect::loopsRemaining() const
  */
 float QSoundEffect::volume() const
 {
-    if (d->m_audioOutput && !d->m_muted)
-        return d->m_audioOutput->volume();
+    if (d->m_audioSink && !d->m_muted)
+        return d->m_audioSink->volume();
 
     return d->m_volume;
 }
@@ -532,7 +631,7 @@ float QSoundEffect::volume() const
 
     UI volume controls should usually be scaled non-linearly. For example, using a logarithmic scale
     will produce linear changes in perceived loudness, which is what a user would normally expect
-    from a volume control. See QAudio::convertVolume() for more details.
+    from a volume control. See QtAudio::convertVolume() for more details.
  */
 void QSoundEffect::setVolume(float volume)
 {
@@ -542,8 +641,8 @@ void QSoundEffect::setVolume(float volume)
 
     d->m_volume = volume;
 
-    if (d->m_audioOutput && !d->m_muted)
-        d->m_audioOutput->setVolume(volume);
+    if (d->m_audioSink && !d->m_muted)
+        d->m_audioSink->setVolume(volume);
 
     emit volumeChanged();
 }
@@ -578,10 +677,10 @@ void QSoundEffect::setMuted(bool muted)
     if (d->m_muted == muted)
         return;
 
-    if (muted && d->m_audioOutput)
-        d->m_audioOutput->setVolume(0);
-    else if (!muted && d->m_audioOutput && d->m_muted)
-        d->m_audioOutput->setVolume(d->m_volume);
+    if (muted && d->m_audioSink)
+        d->m_audioSink->setVolume(0);
+    else if (!muted && d->m_audioSink && d->m_muted)
+        d->m_audioSink->setVolume(d->m_volume);
 
     d->m_muted = muted;
     emit mutedChanged();
