@@ -135,7 +135,8 @@ void QMediaPlayerPrivate::setMedia(const QUrl &media, QIODevice *stream)
     // Back ends can't play qrc files directly.
     // If the back end supports StreamPlayback, we pass a QFile for that resource.
     // If it doesn't, we copy the data to a temporary file and pass its path.
-    if (!media.isEmpty() && !stream && media.scheme() == QLatin1String("qrc")) {
+    if (!media.isEmpty() && !stream && media.scheme() == QLatin1String("qrc")
+        && !control->canPlayQrc()) {
         qrcMedia = media;
 
         file.reset(new QFile(QLatin1Char(':') + media.path()));
