@@ -6,9 +6,11 @@
 #include <private/qplatformmediarecorder_p.h>
 #include <qaudiodevice.h>
 #include <qcamera.h>
+#include <qscreencapture.h>
+#include <qwindowcapture.h>
 #include <qmediacapturesession.h>
 #include <private/qplatformcamera_p.h>
-#include <private/qplatformscreencapture_p.h>
+#include <private/qplatformsurfacecapture_p.h>
 #include <private/qplatformmediaintegration_p.h>
 #include <private/qplatformmediacapture_p.h>
 
@@ -32,14 +34,35 @@ QT_BEGIN_NAMESPACE
 
     \brief The QMediaRecorder class is used for encoding and recording a capture session.
 
-    The QMediaRecorder class is a class for encoding and recording media generated in a
-    QMediaCaptureSession.
+    Use the QMediaRecorder class to encode and record media generated in
+    \l QMediaCaptureSession. You can generate:
+    \list
+        \li Audio. Use \l QAudioInput or \l QAudioBufferInput.
+        \li Video. Use \l QCamera, \l QScreenCapture, \l QWindowCapture, or \l QVideoFrameInput.
+    \endlist
+
+    To record media, connect a generator to a corresponding media capture session.
+
+    Performance of video encoding and recording is limited by the hardware,
+    the operating system, the installed graphic drivers, and the input video format.
+    If \c QCamera, \c QScreenCapture, or \c QWindowCapture produces video frames
+    faster than \c QMediaRecorder can encode and record them, the recorder
+    may drop some frames. This is likely to occur if the input frame resolution
+    is high, 4K for example, and hardware-accelerated encoding is unavailable.
+    If you generate input video via \c QVideoFrameInput, the method
+    \c QVideoFrameInput::sendVideoFrame will do nothing and return \c false
+    whenever this limitation is reached and the internal frame queue is full.
+    Rely on the signal \c QVideoFrameInput::readyToSendVideoFrame to know
+    when the recorder is ready to receive new frames again.
+    If you cannot change the rate of video frame generation and dropping frames
+    is undesirable, we recommend implementing your own frame queue on top of
+    \c QVideoFrameInput, considering the memory limitations of the hardware.
 
     \snippet multimedia-snippets/media.cpp Media recorder
 */
 /*!
     \qmltype MediaRecorder
-    \instantiates QMediaRecorder
+    \nativetype QMediaRecorder
     \brief For encoding and recording media generated in a CaptureSession.
 
     \inqmlmodule QtMultimedia
@@ -47,8 +70,18 @@ QT_BEGIN_NAMESPACE
     \ingroup multimedia_audio_qml
     \ingroup multimedia_video_qml
 
-    The MediaRecorder element can be used within a CaptureSession to record and encode audio and
-    video captured from a microphone and camera
+    Use the MediaRecorder element within a CaptureSession to encode and record:
+    \list
+        \li Audio captured from an audio interface (like microphone or line input).
+        \li Video captured from camera, screen, or an application window.
+    \endlist
+
+    Performance of video encoding and recording is limited by the hardware,
+    the operating system, the installed graphic drivers, and the input video format.
+    If \c Camera, \c ScreenCapture, or \c WindowCapture produces video frames
+    faster than \c MediaRecorder can encode and record them, the recorder
+    may drop some frames. This is likely to occur if the input frame resolution
+    is high, 4K for example, and hardware-accelerated encoding is unavailable.
 
     \since 6.2
     The code below shows a simple capture session containing a MediaRecorder using the default
@@ -84,7 +117,7 @@ QT_BEGIN_NAMESPACE
     }
 \endqml
 
-    \sa CaptureSession, Camera, AudioInput, ImageCapture
+    \sa CaptureSession, Camera, ScreenCapture, WindowCapture, AudioInput, ImageCapture
 */
 QMediaRecorderPrivate::QMediaRecorderPrivate()
 {
@@ -99,7 +132,7 @@ QString QMediaRecorderPrivate::msgFailedStartRecording()
 }
 
 /*!
-    Constructs a media recorder which records the media produced by a microphone and camera.
+    Constructs a media recorder.
     The media recorder is a child of \a{parent}.
 */
 
@@ -108,9 +141,17 @@ QMediaRecorder::QMediaRecorder(QObject *parent)
       d_ptr(new QMediaRecorderPrivate)
 {
     Q_D(QMediaRecorder);
+
+    auto &mediaIntegration = *QPlatformMediaIntegration::instance();
+
     d->q_ptr = this;
-    auto maybeControl = QPlatformMediaIntegration::instance()->createRecorder(this);
+    auto maybeControl = mediaIntegration.createRecorder(this);
     if (maybeControl) {
+        // The first format info initialization may take some time,
+        // for users it seems to be more suitable to have a delay on the object construction
+        // rather than on QMediaRecorder::record
+        mediaIntegration.formatInfo();
+
         d->control = maybeControl.value();
     } else {
         d->initErrorMessage = maybeControl.error();
@@ -152,48 +193,80 @@ void QMediaRecorder::setCaptureSession(QMediaCaptureSession *session)
 
     Setting the location can fail, for example when the service supports only
     local file system locations but a network URL was passed. If the operation
-    fails an errorOccured() signal is emitted.
+    fails an \l errorOccured() signal is emitted.
 
-    The location can be relative or empty. If empty the recorder uses the
-    system specific place and file naming scheme.
+    The output location can be empty, a directory, or a file. The path to a
+    directory or file can be relative or absolute. The \l record() method
+    generates the actual location according to the specified output location and
+    system-specific settings. Refer to the \l actualLocation property description
+    for details.
 
-   \sa errorOccurred()
+   \sa actualLocation, errorOccurred()
 */
 
 /*!
     \property QMediaRecorder::outputLocation
-    \brief the destination location of media content.
+    \brief The destination location of media content.
 
     Setting the location can fail, for example when the service supports only
     local file system locations but a network URL was passed. If the operation
-    fails an errorOccured() signal is emitted.
+    fails, the \l errorOccured() signal is emitted.
 
-    The output location can be relative or empty; in the latter case the recorder
-    uses the system specific place and file naming scheme.
+    The output location is ignored if a writable \l outputDevice
+    has been assigned to the recorder.
+    This behavior may change in the future, so we recommend setting only one output,
+    either \c outputLocation or \c outputDevice.
+
+    The output location can be empty, a directory, or a file. The path to a
+    directory or file can be relative or absolute. The \l record() method
+    generates the actual location according to the specified output location and
+    system-specific settings. Refer to the \l actualLocation property description
+    for details.
+
+    \sa actualLocation, outputDevice()
 */
 
 /*!
     \qmlproperty QUrl QtMultimedia::MediaRecorder::actualLocation
     \brief The actual location of the last media content.
 
-    The actual location is usually available after recording starts,
-    and reset when new location is set or new recording starts.
+    The actual location is reset when a new \l outputLocation is assigned.
+    When \l record() is invoked, the recorder generates the actual location
+    basing on the following rules.
+    \list
+        \li If \c outputLocation is empty, a directory, or a file
+            without an extension, the recorder generates the appropriate extension
+            based on the selected media format and system MIME types.
+        \li If \c outputLocation is a directory, the recorder generates a new file
+            name within it.
+        \li If \c outputLocation is empty, the recorder generates a new file name in
+            the system-specific directory for audio or video.
+        \li The recorder generates the actual location before
+            emitting \c recorderStateChanged(RecordingState).
+    \endlist
 */
 
 /*!
     \property QMediaRecorder::actualLocation
     \brief The actual location of the last media content.
 
-    The actual location is usually available after recording starts,
-    and reset when new location is set or new recording starts.
+    The actual location is reset when a new \l outputLocation
+    or a non-null \l outputDevice is assigned.
+    When \l record() is invoked and \c outputDevice is \c null or not writable,
+    the recorder generates the actual location basing on the following rules.
+    \list
+        \li If \c outputLocation is empty, a directory, or a file
+            without an extension, the recorder generates the appropriate extension
+            based on the selected media format and system MIME types.
+        \li If \c outputLocation is a directory, the recorder generates a new file
+            name within it.
+        \li If \c outputLocation is empty, the recorder generates a new file name in
+            the system-specific directory for audio or video.
+        \li The recorder generates the actual location before
+            emitting \c recorderStateChanged(RecordingState).
+    \endlist
 */
 
-/*!
-    \qmlproperty bool QtMultimedia::MediaRecorder::isAvailable
-    \brief This property holds whether the recorder service is ready to use.
-
-    Returns \c true if media recorder service ready to use.
-*/
 /*!
     Returns \c true if media recorder service ready to use.
 */
@@ -217,7 +290,55 @@ void QMediaRecorder::setOutputLocation(const QUrl &location)
     d->control->setOutputLocation(location);
     d->control->clearActualLocation();
     if (!location.isEmpty() && !d->control->isLocationWritable(location))
-            emit errorOccurred(QMediaRecorder::LocationNotWritable, tr("Output location not writable"));
+        emit errorOccurred(QMediaRecorder::LocationNotWritable,
+                           QStringLiteral("Output location not writable"));
+}
+
+/*!
+    Set the output IO device for media content.
+
+    The \a device must have been opened in the \l{QIODevice::WriteOnly}{WriteOnly} or
+    \l{QIODevice::ReadWrite}{ReadWrite} modes before the recording starts.
+
+    The media recorder doesn't take ownership of the specified \a device.
+    If the recording has been started, the device must be kept alive and open until
+    the signal \c recorderStateChanged(StoppedState) is emitted.
+
+    This method resets \l actualLocation immediately unless
+    the specified \a device is \c null.
+
+    If a writable output device is assigned to the recorder, \l outputLocation
+    is ignored, and \l actualLocation is not generated when recording starts.
+    This behavior may change in the future, so we recommend setting only
+    one output, either \c outputLocation or \c outputDevice.
+
+    \c QMediaRecorder::setOutputDevice is only supported with the FFmpeg backend.
+
+    \sa outputDevice(), outputLocation
+*/
+void QMediaRecorder::setOutputDevice(QIODevice *device)
+{
+    Q_D(QMediaRecorder);
+    if (!d->control) {
+        emit errorOccurred(QMediaRecorder::ResourceError, d->initErrorMessage);
+        return;
+    }
+
+    d->control->setOutputDevice(device);
+
+    if (device)
+        d->control->clearActualLocation();
+}
+
+/*!
+    Returns the output IO device for media content.
+
+    \sa setOutputDevice()
+*/
+QIODevice *QMediaRecorder::outputDevice() const
+{
+    Q_D(const QMediaRecorder);
+    return d->control ? d->control->outputDevice() : nullptr;
 }
 
 QUrl QMediaRecorder::actualLocation() const
@@ -302,6 +423,8 @@ qint64 QMediaRecorder::duration() const
     If recording fails, the error() signal is emitted with recorder state being
     reset back to \c{QMediaRecorder.StoppedState}.
 
+    This method updates \l actualLocation according to its generation rules.
+
     \note On mobile devices, recording will happen in the orientation the
     device had when calling record and is locked for the duration of the recording.
     To avoid artifacts on the user interface, we recommend to keep the user interface
@@ -310,7 +433,7 @@ qint64 QMediaRecorder::duration() const
     is finished.
 */
 /*!
-    Start recording.
+    Starts recording.
 
     While the recorder state is changed immediately to
     c\{QMediaRecorder::RecordingState}, recording may start asynchronously.
@@ -318,6 +441,8 @@ qint64 QMediaRecorder::duration() const
     If recording fails error() signal is emitted with recorder state being
     reset back to \c{QMediaRecorder::StoppedState}.
 
+    This method updates \l actualLocation according to its generation rules.
+
     \note On mobile devices, recording will happen in the orientation the
     device had when calling record and is locked for the duration of the recording.
     To avoid artifacts on the user interface, we recommend to keep the user interface
@@ -330,16 +455,16 @@ void QMediaRecorder::record()
 {
     Q_D(QMediaRecorder);
 
-    if (!d->control || ! d->captureSession)
+    if (!d->control || !d->captureSession)
         return;
 
     if (d->control->state() == QMediaRecorder::PausedState) {
         d->control->resume();
     } else {
         auto oldMediaFormat = d->encoderSettings.mediaFormat();
-        auto camera = d->captureSession->camera();
-        auto screenCapture = d->captureSession->screenCapture();
-        bool hasVideo = (camera && camera->isActive()) || (screenCapture && screenCapture->isActive());
+
+        auto platformSession = d->captureSession->platformSession();
+        const bool hasVideo = platformSession && !platformSession->activeVideoSources().empty();
 
         d->encoderSettings.resolveFormat(hasVideo ? QMediaFormat::RequiresVideo : QMediaFormat::NoFlags);
         d->control->clearActualLocation();
@@ -540,10 +665,52 @@ void QMediaRecorder::addMetaData(const QMediaMetaData &metaData)
 {
     auto data = this->metaData();
     // merge data
-    for (const auto &k : metaData.keys())
-        data.insert(k, metaData.value(k));
+    for (auto &&[key, value] : metaData.asKeyValueRange())
+        data.insert(key, value);
     setMetaData(data);
 }
+
+/*!
+    \property QMediaRecorder::autoStop
+
+    This property controls whether the media recorder stops automatically when
+    all media inputs have reported the end of the stream or have been deactivated.
+
+    The end of the stream is reported by sending an empty media frame,
+    which you can send explicitly via \l QVideoFrameInput or \l QAudioBufferInput.
+
+    Video inputs, specificly, \l QCamera, \l QScreenCapture and \l QWindowCapture,
+    can be deactivated via the function \c setActive.
+
+    Defaults to \c false.
+
+    QMediaRecorder::autoStop is only supported with the FFmpeg backend.
+
+    \sa QCamera, QScreenCapture, QWindowCapture
+*/
+
+bool QMediaRecorder::autoStop() const
+{
+    Q_D(const QMediaRecorder);
+
+    return d->autoStop;
+}
+
+void QMediaRecorder::setAutoStop(bool autoStop)
+{
+    Q_D(QMediaRecorder);
+
+    if (d->autoStop == autoStop)
+        return;
+
+    d->autoStop = autoStop;
+
+    if (d->control)
+        d->control->updateAutoStop();
+
+    emit autoStopChanged();
+}
+
 /*!
     \qmlsignal QtMultimedia::MediaRecorder::metaDataChanged()
 
@@ -632,6 +799,14 @@ void QMediaRecorder::setMediaFormat(const QMediaFormat &format)
     emit mediaFormatChanged();
 }
 
+/*!
+
+    \qmlproperty enumeration QtMultimedia::MediaRecorder::encodingMode
+    \since 6.6
+    \brief This property holds the encoding mode.
+    \sa QMediaRecorder::EncodingMode
+*/
+
 /*!
     Returns the encoding mode.
 
@@ -691,6 +866,15 @@ void QMediaRecorder::setQuality(Quality quality)
     emit qualityChanged();
 }
 
+/*!
+    \qmlproperty Size QtMultimedia::MediaRecorder::videoResolution
+    \since 6.6
+    \brief This property holds the resolution of the encoded video.
+
+    Set an empty Size to make the recorder choose an optimal resolution based
+    on what is available from the video source and the limitations of the codec.
+*/
+
 
 /*!
     Returns the resolution of the encoded video.
@@ -728,6 +912,15 @@ void QMediaRecorder::setVideoResolution(const QSize &size)
     \overload
 */
 
+/*!
+    \qmlproperty real QtMultimedia::MediaRecorder::videoFrameRate
+    \since 6.6
+    \brief This property holds the video frame rate.
+
+    A value of 0 indicates the recorder should make an optimal choice based on what is available
+    from the video source and the limitations of the codec.
+*/
+
 /*!
     Returns the video frame rate.
 */
@@ -757,6 +950,12 @@ void QMediaRecorder::setVideoFrameRate(qreal frameRate)
     emit videoFrameRateChanged();
 }
 
+/*!
+    \qmlproperty int QtMultimedia::MediaRecorder::videoBitRate
+    \since 6.6
+    \brief This property holds the bit rate of the compressed video stream in bits per second.
+*/
+
 /*!
     Returns the bit rate of the compressed video stream in bits per second.
 */
@@ -783,6 +982,12 @@ void QMediaRecorder::setVideoBitRate(int bitRate)
     emit videoBitRateChanged();
 }
 
+/*!
+    \qmlproperty int QtMultimedia::MediaRecorder::audioBitRate
+    \since 6.6
+    \brief This property holds the bit rate of the compressed audio stream in bits per second.
+*/
+
 /*!
     Returns the bit rate of the compressed audio stream in bits per second.
 */
@@ -809,6 +1014,12 @@ void QMediaRecorder::setAudioBitRate(int bitRate)
     emit audioBitRateChanged();
 }
 
+/*!
+    \qmlproperty int QtMultimedia::MediaRecorder::audioChannelCount
+    \since 6.6
+    \brief This property holds the number of audio channels.
+*/
+
 /*!
     Returns the number of audio channels.
 */
@@ -838,6 +1049,12 @@ void QMediaRecorder::setAudioChannelCount(int channels)
     emit audioChannelCountChanged();
 }
 
+/*!
+    \qmlproperty int QtMultimedia::MediaRecorder::audioSampleRate
+    \since 6.6
+    \brief This property holds the audio sample rate in Hz.
+*/
+
 /*!
     Returns the audio sample rate in Hz.
 */
