@@ -24,7 +24,7 @@
 #include <QtCore/qurl.h>
 
 #include <common/qgst_p.h>
-#include <common/qgst_bus_p.h>
+#include <common/qgst_bus_observer_p.h>
 #include <common/qgstpipeline_p.h>
 
 #include <gst/app/gstappsink.h>
@@ -61,6 +61,8 @@ public:
     // GStreamerBusMessageFilter interface
     bool processBusMessage(const QGstreamerMessage &message) override;
 
+    bool canReadQrc() const override;
+
 private slots:
     void updateDuration();
 
@@ -70,9 +72,6 @@ private:
     static GstFlowReturn new_sample(GstAppSink *sink, gpointer user_data);
     GstFlowReturn newSample(GstAppSink *sink);
 
-    static void configureAppSrcElement(GObject *, GObject *, GParamSpec *,
-                                       QGstreamerAudioDecoder *_this);
-
     void setAudioFlags(bool wantNativeAudio);
     void addAppSink();
     void removeAppSink();
